@@ -23,20 +23,20 @@
 #include <QImageReader>
 #include <QMessageBox>
 #include <QPixmap>
+#include <QRect>
 #include <QStringList>
+#include <QTime>
 #include <QX11Info>
 
-#include "constants.h"
 #include "trayitem.h"
-#include "util.h"
+#include "xlibutil.h"
 
 #include <Xatom.h>
 #include <X11/xpm.h>
 
 #include "mwmutil.h"
 
-TrayItem::TrayItem(Window window, QObject *parent) : QSystemTrayIcon(parent) {
-    m_iconified = false;
+TrayItem::TrayItem(Window window) {
     m_customIcon = false;
     m_skipTaskbar = false;
     m_skipPager = false;
@@ -44,15 +44,76 @@ TrayItem::TrayItem(Window window, QObject *parent) : QSystemTrayIcon(parent) {
     m_iconifyMinimized = true;
     m_iconifyObscure = false;
     m_iconifyFocusLost = false;
+    m_iconifyOnClose = true;
     m_balloonTimeout = 4000;
-    m_window = window;
     m_dockedAppName = "";
+    m_window = window;
 
     Display *display = QX11Info::display();
-    // Allows events from m_window to be forwarded to the x11EventFilter.
-    subscribe(display, m_window, StructureNotifyMask | PropertyChangeMask | VisibilityChangeMask | FocusChangeMask, true);
+    long dummy;
+    Window root;
+    int x;
+    int y;
+    unsigned int width;
+    unsigned int height;
+    unsigned int border;
+    unsigned int depth;
+    Atom wm_hints_atom = XInternAtom(display, _XA_MOTIF_WM_HINTS, false);
+    unsigned char *wm_data;
+    Atom wm_type;
+    int wm_format;
+    unsigned long wm_nitems;
+    unsigned long wm_bytes_after;
+
+    // Get the window size info.
+    XGetWMNormalHints(display, m_window, &m_sizeHint, &dummy);
+    XGetGeometry(display, m_window, &root, &x, &y, &width, &height, &border, &depth);
+    // Get the window decoration info.
+    XGetWindowProperty(display, m_window, wm_hints_atom, 0, sizeof (MotifWmHints) / sizeof (long), false, AnyPropertyType, &wm_type, &wm_format, &wm_nitems, &wm_bytes_after, &wm_data);
+
+    // There is a conflict between QX11EmbedContainer and QSystemTrayIcon with
+    // regard to registering the X11EventFilter. This prevents a segfault.
+    QX11EmbedContainer();
+    // Create the container window and place the selected window into it.
+    m_container = new QX11EmbedContainer();
+    m_container->embedClient(m_window);
+    m_container->show();
+
+    // Allows events from m_container and m_window to be forwarded to the
+    // x11EventFilter. Only event types that are subscribed to will be sent on.
+    XLibUtil::subscribe(display, m_container->winId(), StructureNotifyMask | PropertyChangeMask | VisibilityChangeMask | FocusChangeMask, true);
+    XLibUtil::subscribe(display, m_window, PropertyChangeMask, true);
+
     // store the desktop on which the window is being shown
-    getCardinalProperty(display, m_window, XInternAtom(display, "_NET_WM_DESKTOP", True), &m_desktop);
+    XLibUtil::getCardinalProperty(display, m_container->winId(), XInternAtom(display, "_NET_WM_DESKTOP", true), &m_desktop);
+
+    // Do not close the window when the X button is pressed. This only works
+    // Because we have embedded into our own window.
+    Atom wm_delete = XInternAtom(display, "WM_DELETE_WINDOW", false);
+    XSetWMProtocols(display, m_container->winId(), &wm_delete, 1);
+
+    // Set the containers size hints to that of window.
+    m_container->setMinimumSize(m_sizeHint.min_width, m_sizeHint.min_height);
+    // Set the containers geometry to that of window.
+    m_container->setGeometry(x, y, width, height);
+
+    // Set the containers decorations to those of window.
+    if (wm_type == None) {
+        MotifWmHints hints;
+        memset(&hints, 0, sizeof (hints));
+        hints.flags = MWM_HINTS_DECORATIONS;
+        hints.decorations = MWM_DECOR_ALL;
+        XChangeProperty(display, m_container->winId(), wm_hints_atom, wm_hints_atom, 32, PropModeReplace, reinterpret_cast<unsigned char *> (& hints), sizeof (MotifWmHints) / sizeof (long));
+    } else {
+        MotifWmHints *hints;
+        hints = reinterpret_cast<MotifWmHints *> (wm_data);
+        if (!(hints->flags & MWM_HINTS_DECORATIONS)) {
+            hints->flags |= MWM_HINTS_DECORATIONS;
+            hints->decorations = 0;
+        }
+        XChangeProperty(display, m_container->winId(), wm_hints_atom, wm_hints_atom, 32, PropModeReplace, reinterpret_cast<unsigned char *> (hints), sizeof (MotifWmHints) / sizeof (long));
+    }
+    XFree(wm_data);
 
     readDockedAppName();
     updateTitle();
@@ -68,168 +129,103 @@ TrayItem::~TrayItem() {
     // Only the main menu needs to be deleted. The rest of the menus and actions
     // are children of this menu and Qt will delete all children.
     delete m_contextMenu;
+
+    // Move and resize the embedded window if it is still open to the size and
+    // position of the container.
+    QRect geo = m_container->geometry();
+    m_container->discardClient();
+    delete m_container;
+    if (m_window) {
+        XMoveResizeWindow(QX11Info::display(), m_window, geo.x(), geo.y(), static_cast<unsigned int> (geo.width()), static_cast<unsigned int> (geo.height()));
+    }
 }
 
-Window TrayItem::dockedWindow() {
+Window TrayItem::containerWindow() {
+    return m_container->winId();
+}
+
+Window TrayItem::embedWindow() {
     return m_window;
 }
 
 bool TrayItem::x11EventFilter(XEvent *ev) {
-    if (isBadWindow()) {
-        return false;
-    }
+    XAnyEvent *event = reinterpret_cast<XAnyEvent *> (ev);
+
+    if (event->window == m_container->winId()) {
+        if (ev->type == ClientMessage) {
+            if (static_cast<unsigned long> (ev->xclient.data.l[0]) == XInternAtom(QX11Info::display(), "WM_DELETE_WINDOW", false)) {
+                if (m_iconifyOnClose) {
+                    iconifyWindow();
+                } else {
+                    closeWindow();
+                }
+                return true;
+            }
+        }
 
-    XAnyEvent *event = (XAnyEvent *) ev;
-    if (event->window == m_window) {
         if (event->type == DestroyNotify) {
             destroyEvent();
+            return true;
         } else if (event->type == PropertyNotify) {
-            propertyChangeEvent(((XPropertyEvent *) event)->atom);
+            return propertyChangeEvent(reinterpret_cast<XPropertyEvent *> (event)->atom);
         } else if (event->type == VisibilityNotify) {
-            if (((XVisibilityEvent *) event)->state == VisibilityFullyObscured) {
+            if (reinterpret_cast<XVisibilityEvent *> (event)->state == VisibilityFullyObscured) {
                 obscureEvent();
+                return true;
             }
         } else if (event->type == FocusOut) {
             focusLostEvent();
-        } else if (event->type == MapNotify) {
-            m_iconified = false;
-            updateToggleAction();
-        } else if (event->type == UnmapNotify) {
-            m_iconified = true;
-            updateToggleAction();
+            return true;
+        }
+    } else if (event->window == m_window) {
+        if (event->type == PropertyNotify) {
+            return updateEmbedProperty(reinterpret_cast<XPropertyEvent *> (event)->atom);
         }
-        return true; // Dont process this again
     }
     return false;
 }
 
 void TrayItem::restoreWindow() {
-    if (isBadWindow()) {
-        return;
-    }
+    m_container->show();
 
     Display *display = QX11Info::display();
     Window root = QX11Info::appRootWindow();
-
-    if (m_iconified) {
-        m_iconified = false;
-        /*
-         * A simple XMapWindow would not do. Some applications like xmms wont
-         * redisplay its other windows (like the playlist, equalizer) since the
-         * Withdrawn->Normal state change code does not map them. So we make the
-         * window go through Withdrawn->Map->Iconify->Normal state.
-         */
-        XMapWindow(display, m_window);
-        XIconifyWindow(display, m_window, DefaultScreen(display));
-        XSync(display, False);
-        long l_state[1] = {NormalState};
-        sendMessage(display, root, m_window, "WM_CHANGE_STATE", 32, SubstructureNotifyMask | SubstructureRedirectMask, l_state, sizeof (l_state));
-
-        m_sizeHint.flags = USPosition;
-        XSetWMNormalHints(display, m_window, &m_sizeHint);
-
-        updateToggleAction();
-    } else {
-        XRaiseWindow(display, m_window);
-    }
+    XRaiseWindow(display, m_container->winId());
 
     // Change to the desktop that the window was last on.
     long l_currDesk[2] = {m_desktop, CurrentTime};
-    sendMessage(display, root, root, "_NET_CURRENT_DESKTOP", 32, SubstructureNotifyMask | SubstructureRedirectMask, l_currDesk, sizeof (l_currDesk));
+    XLibUtil::sendMessage(display, root, root, "_NET_CURRENT_DESKTOP", 32, SubstructureNotifyMask | SubstructureRedirectMask, l_currDesk, sizeof (l_currDesk));
     // Set the desktop the window wants to be on.
     long l_wmDesk[2] = {m_desktop, 1}; // 1 == request sent from application. 2 == from pager
-    sendMessage(display, root, m_window, "_NET_WM_DESKTOP", 32, SubstructureNotifyMask | SubstructureRedirectMask, l_wmDesk, sizeof (l_wmDesk));
+    XLibUtil::sendMessage(display, root, m_container->winId(), "_NET_WM_DESKTOP", 32, SubstructureNotifyMask | SubstructureRedirectMask, l_wmDesk, sizeof (l_wmDesk));
 
-    // Make it the active window
-    long l_active[2] = {1, CurrentTime}; // 1 == request sent from application. 2 == from pager
-    sendMessage(display, root, m_window, "_NET_ACTIVE_WINDOW", 32, SubstructureNotifyMask | SubstructureRedirectMask, l_active, sizeof (l_active));
-    XSetInputFocus(display, m_window, RevertToParent, CurrentTime);
+    m_container->activateWindow();
+    updateToggleAction();
+    doSkipTaskbar();
 }
 
 void TrayItem::iconifyWindow() {
-    if (isBadWindow()) {
-        return;
-    }
-
-    m_iconified = true;
-
-    Display *display = QX11Info::display();
-    int screen = DefaultScreen(display);
-    long dummy;
-
-    XGetWMNormalHints(display, m_window, &m_sizeHint, &dummy);
-
-    /*
-     * A simple call to XWithdrawWindow wont do. Here is what we do:
-     * 1. Iconify. This will make the application hide all its other windows. For
-     *    example, xmms would take off the playlist and equalizer window.
-     * 2. Withdrawn the window so we will be removed from the taskbar.
-     */
-    XIconifyWindow(display, m_window, screen); // good for effects too
-    XSync(display, False);
-    XWithdrawWindow(display, m_window, screen);
+    m_container->hide();
     updateToggleAction();
 }
 
-void TrayItem::skip_NET_WM_STATE(const char *type, bool set) {
-    // set, true = add the state to the window. False, remove the state from
-    // the window.
-
-    if (isBadWindow() || m_iconified) {
-        return;
-    }
+void TrayItem::closeWindow() {
     Display *display = QX11Info::display();
-    Atom atom = XInternAtom(display, type, False);
-
-    long l[5] = {set ? 1 : 0, atom, 0, 0, 0};
-    sendMessage(display, QX11Info::appRootWindow(), m_window, "_NET_WM_STATE", 32, SubstructureNotifyMask, l, sizeof (l));
+    long l[2] = {XInternAtom(QX11Info::display(), "WM_DELETE_WINDOW", false), CurrentTime};
+    restoreWindow();
+    XLibUtil::sendMessage(display, m_window, m_window, "WM_PROTOCOLS", 32, NoEventMask, l, sizeof (l));
 }
 
-void TrayItem::skipTaskbar() {
-    skip_NET_WM_STATE("_NET_WM_STATE_SKIP_TASKBAR", m_skipTaskbar);
+void TrayItem::doSkipTaskbar() {
+    set_NET_WM_STATE("_NET_WM_STATE_SKIP_TASKBAR", m_skipTaskbar);
 }
 
-void TrayItem::skipPager() {
-    skip_NET_WM_STATE("_NET_WM_STATE_SKIP_PAGER", m_skipPager);
+void TrayItem::doSkipPager() {
+    set_NET_WM_STATE("_NET_WM_STATE_SKIP_PAGER", m_skipPager);
 }
 
-void TrayItem::sticky() {
-    skip_NET_WM_STATE("_NET_WM_STATE_STICKY", m_sticky);
-}
-
-void TrayItem::removeWindowBorder() {
-    if (isBadWindow()) {
-        return;
-    }
-
-    Display *display = QX11Info::display();
-    Atom hints_atom = XInternAtom(display, _XA_MOTIF_WM_HINTS, false);
-
-    unsigned char *data;
-    Atom type;
-    int format;
-    unsigned long nitems;
-    unsigned long bytes_after;
-
-    XGetWindowProperty(display, m_window, hints_atom, 0, sizeof (MotifWmHints) / sizeof (long), false, AnyPropertyType, &type, &format, &nitems, &bytes_after, &data);
-
-    if (type == None) {
-        MotifWmHints hints;
-        memset(&hints, 0, sizeof (hints));
-        hints.flags = MWM_HINTS_DECORATIONS;
-        hints.decorations = 0;
-        XChangeProperty(display, m_window, hints_atom, hints_atom, 32, PropModeReplace, (unsigned char *) & hints, sizeof (MotifWmHints) / sizeof (long));
-    } else {
-        MotifWmHints *hints;
-        hints = (MotifWmHints *) data;
-
-        if (!(hints->flags & MWM_HINTS_DECORATIONS)) {
-            hints->flags |= MWM_HINTS_DECORATIONS;
-        }
-        hints->decorations = 0;
-        XChangeProperty(display, m_window, hints_atom, hints_atom, 32, PropModeReplace, (unsigned char *) hints, sizeof (MotifWmHints) / sizeof (long));
-    }
-    XFree(data);
+void TrayItem::doSticky() {
+    set_NET_WM_STATE("_NET_WM_STATE_STICKY", m_sticky);
 }
 
 void TrayItem::setCustomIcon(QString path) {
@@ -242,18 +238,6 @@ void TrayItem::setCustomIcon(QString path) {
     setIcon(QIcon(customIcon));
 }
 
-void TrayItem::close() {
-    if (isBadWindow()) {
-        return;
-    } else {
-        Display *display = QX11Info::display();
-        long l[5] = {0, 0, 0, 0, 0};
-        restoreWindow();
-        sendMessage(display, QX11Info::appRootWindow(), m_window, "_NET_CLOSE_WINDOW", 32, SubstructureNotifyMask | SubstructureRedirectMask, l, sizeof (l));
-        destroyEvent();
-    }
-}
-
 void TrayItem::selectCustomIcon(bool value) {
     Q_UNUSED(value);
 
@@ -280,19 +264,19 @@ void TrayItem::selectCustomIcon(bool value) {
 void TrayItem::setSkipTaskbar(bool value) {
     m_skipTaskbar = value;
     m_actionSkipTaskbar->setChecked(value);
-    skipTaskbar();
+    doSkipTaskbar();
 }
 
 void TrayItem::setSkipPager(bool value) {
     m_skipPager = value;
     m_actionSkipPager->setChecked(value);
-    skipPager();
+    doSkipPager();
 }
 
 void TrayItem::setSticky(bool value) {
     m_sticky = value;
     m_actionSticky->setChecked(value);
-    sticky();
+    doSticky();
 }
 
 void TrayItem::setIconifyMinimized(bool value) {
@@ -311,6 +295,11 @@ void TrayItem::setIconifyFocusLost(bool value) {
     focusLostEvent();
 }
 
+void TrayItem::setIconifyOnClose(bool value) {
+    m_iconifyOnClose = value;
+    m_actionIconifyOnClose->setChecked(value);
+}
+
 void TrayItem::setBalloonTimeout(int value) {
     if (value < 0) {
         value = 0;
@@ -328,31 +317,25 @@ void TrayItem::setBalloonTimeout(bool value) {
 }
 
 void TrayItem::toggleWindow() {
-    if (m_iconified) {
-        restoreWindow();
-    } else {
+    if (m_container->isVisible()) {
         iconifyWindow();
+    } else {
+        restoreWindow();
     }
 }
 
 void TrayItem::trayActivated(QSystemTrayIcon::ActivationReason reason) {
     if (reason == QSystemTrayIcon::Trigger) {
-        if (m_iconified || m_window != activeWindow(QX11Info::display())) {
-            restoreWindow();
-        } else {
+        if (m_container->winId() == XLibUtil::activeWindow(QX11Info::display())) {
             iconifyWindow();
+        } else {
+            restoreWindow();
         }
     }
 }
 
 void TrayItem::doAbout() {
-    QMessageBox aboutBox;
-    aboutBox.setIconPixmap(QPixmap(":/images/kdocker.png"));
-    aboutBox.setWindowTitle(tr("About %1 - %2").arg(qApp->applicationName()).arg(qApp->applicationVersion()));
-    aboutBox.setText(ABOUT);
-    aboutBox.setInformativeText(tr("See %1 for more information.").arg("<a href=\"https://launchpad.net/kdocker\">https://launchpad.net/kdocker</a>"));
-    aboutBox.setStandardButtons(QMessageBox::Ok);
-    aboutBox.exec();
+    emit(about());
 }
 
 void TrayItem::doSelectAnother() {
@@ -378,36 +361,42 @@ void TrayItem::destroyEvent() {
     emit(dead(this));
 }
 
-void TrayItem::propertyChangeEvent(Atom property) {
-    if (isBadWindow()) {
-        return;
-    }
-
+bool TrayItem::propertyChangeEvent(Atom property) {
     Display *display = QX11Info::display();
-    static Atom WM_NAME = XInternAtom(display, "WM_NAME", True);
-    static Atom WM_ICON = XInternAtom(display, "WM_ICON", True);
     static Atom WM_STATE = XInternAtom(display, "WM_STATE", True);
     static Atom _NET_WM_DESKTOP = XInternAtom(display, "_NET_WM_DESKTOP", True);
 
-    if (property == WM_NAME) {
-        updateTitle();
-    } else if (property == WM_ICON) {
-        updateIcon();
-    } else if (property == _NET_WM_DESKTOP) {
-        getCardinalProperty(display, m_window, _NET_WM_DESKTOP, &m_desktop);
+    if (property == _NET_WM_DESKTOP) {
+        XLibUtil::getCardinalProperty(display, m_container->winId(), _NET_WM_DESKTOP, &m_desktop);
+        return true;
     } else if (property == WM_STATE) {
         Atom type = None;
         int format;
         unsigned long nitems, after;
         unsigned char *data = 0;
-        int r = XGetWindowProperty(display, m_window, WM_STATE, 0, 1, False, AnyPropertyType, &type, &format, &nitems, &after, &data);
-        if ((r == Success) && data && (*(long *) data == IconicState)) {
+        int r = XGetWindowProperty(display, m_container->winId(), WM_STATE, 0, 1, False, AnyPropertyType, &type, &format, &nitems, &after, &data);
+        if ((r == Success) && data && (*reinterpret_cast<long *> (data) == IconicState)) {
             minimizeEvent();
             XFree(data);
+            return true;
         }
-        skipTaskbar();
-        skipPager();
     }
+    return false;
+}
+
+bool TrayItem::updateEmbedProperty(Atom property) {
+    Display *display = QX11Info::display();
+    static Atom WM_NAME = XInternAtom(display, "WM_NAME", True);
+    static Atom WM_ICON = XInternAtom(display, "WM_ICON", True);
+
+    if (property == WM_NAME) {
+        updateTitle();
+        return true;
+    } else if (property == WM_ICON) {
+        updateIcon();
+        return true;
+    }
+    return false;
 }
 
 void TrayItem::obscureEvent() {
@@ -417,16 +406,30 @@ void TrayItem::obscureEvent() {
 }
 
 void TrayItem::focusLostEvent() {
-    if (m_iconifyFocusLost && m_window != activeWindow(QX11Info::display())) {
+    // Wait half a second before checking to ensure the window is properly
+    // focused when being restored.
+    QTime t;
+    t.start();
+    while (t.elapsed() < 500) {
+        qApp->processEvents();
+    }
+
+    if (m_iconifyFocusLost && m_container->winId() != XLibUtil::activeWindow(QX11Info::display())) {
         iconifyWindow();
     }
 }
 
-void TrayItem::readDockedAppName() {
-    if (isBadWindow()) {
-        return;
-    }
+void TrayItem::set_NET_WM_STATE(const char *type, bool set) {
+    // set, true = add the state to the window. False, remove the state from
+    // the window.
+    Display *display = QX11Info::display();
+    Atom atom = XInternAtom(display, type, False);
 
+    long l[2] = {set ? 1 : 0, atom};
+    XLibUtil::sendMessage(display, QX11Info::appRootWindow(), m_container->winId(), "_NET_WM_STATE", 32, SubstructureNotifyMask, l, sizeof (l));
+}
+
+void TrayItem::readDockedAppName() {
     Display *display = QX11Info::display();
     XClassHint ch;
     if (XGetClassHint(display, m_window, &ch)) {
@@ -449,10 +452,6 @@ void TrayItem::readDockedAppName() {
  * Update the title in the tooltip.
  */
 void TrayItem::updateTitle() {
-    if (isBadWindow()) {
-        return;
-    }
-
     Display *display = QX11Info::display();
     char *windowName = 0;
     QString title;
@@ -463,6 +462,8 @@ void TrayItem::updateTitle() {
         XFree(windowName);
     }
 
+    m_container->setWindowTitle(title);
+
     setToolTip(QString("%1 [%2]").arg(title).arg(m_dockedAppName));
     if (m_balloonTimeout > 0) {
         showMessage(m_dockedAppName, title, QSystemTrayIcon::Information, m_balloonTimeout);
@@ -470,23 +471,28 @@ void TrayItem::updateTitle() {
 }
 
 void TrayItem::updateIcon() {
-    if (isBadWindow() || m_customIcon) {
+    if (m_customIcon) {
         return;
     }
 
-    setIcon(createIcon(m_window));
+    QIcon icon = createIcon(m_window);
+
+    m_container->setWindowIcon(icon);
+    setIcon(icon);
 }
 
 void TrayItem::updateToggleAction() {
     QString text;
     QIcon icon;
-    if (m_iconified) {
-        text = tr("Show %1").arg(m_dockedAppName);
-        icon = QIcon(":/images/restore.png");
-    } else {
+
+    if (m_container->isVisible()) {
         text = tr("Hide %1").arg(m_dockedAppName);
         icon = QIcon(":/images/iconify.png");
+    } else {
+        text = tr("Show %1").arg(m_dockedAppName);
+        icon = QIcon(":/images/restore.png");
     }
+
     m_actionToggle->setIcon(icon);
     m_actionToggle->setText(text);
 }
@@ -499,46 +505,60 @@ void TrayItem::createContextMenu() {
 
     // Options menu
     m_optionsMenu = new QMenu(tr("Options"), m_contextMenu);
+
     m_actionSetIcon = new QAction(tr("Set icon"), m_optionsMenu);
     connect(m_actionSetIcon, SIGNAL(triggered(bool)), this, SLOT(selectCustomIcon(bool)));
     m_optionsMenu->addAction(m_actionSetIcon);
+
     m_actionSkipTaskbar = new QAction(tr("Skip taskbar"), m_optionsMenu);
     m_actionSkipTaskbar->setCheckable(true);
     m_actionSkipTaskbar->setChecked(m_skipTaskbar);
     connect(m_actionSkipTaskbar, SIGNAL(triggered(bool)), this, SLOT(setSkipTaskbar(bool)));
     m_optionsMenu->addAction(m_actionSkipTaskbar);
+
     m_actionSkipPager = new QAction(tr("Skip pager"), m_optionsMenu);
     m_actionSkipPager->setCheckable(true);
     m_actionSkipPager->setChecked(m_skipPager);
     connect(m_actionSkipPager, SIGNAL(triggered(bool)), this, SLOT(setSkipPager(bool)));
     m_optionsMenu->addAction(m_actionSkipPager);
+
     m_actionSticky = new QAction(tr("Sticky"), m_optionsMenu);
     m_actionSticky->setCheckable(true);
     m_actionSticky->setChecked(m_sticky);
     connect(m_actionSticky, SIGNAL(triggered(bool)), this, SLOT(setSticky(bool)));
     m_optionsMenu->addAction(m_actionSticky);
+
     m_actionIconifyMinimized = new QAction(tr("Iconify when minimized"), m_optionsMenu);
     m_actionIconifyMinimized->setCheckable(true);
     m_actionIconifyMinimized->setChecked(m_iconifyMinimized);
     connect(m_actionIconifyMinimized, SIGNAL(triggered(bool)), this, SLOT(setIconifyMinimized(bool)));
     m_optionsMenu->addAction(m_actionIconifyMinimized);
+
     m_actionIconifyObscure = new QAction(tr("Iconify when obscured"), m_optionsMenu);
     m_actionIconifyObscure->setCheckable(true);
     m_actionIconifyObscure->setChecked(m_iconifyObscure);
     connect(m_actionIconifyObscure, SIGNAL(triggered(bool)), this, SLOT(setIconifyObscure(bool)));
     m_optionsMenu->addAction(m_actionIconifyObscure);
+
     m_actionIconifyFocusLost = new QAction(tr("Iconify when focus lost"), m_optionsMenu);
     m_actionIconifyFocusLost->setCheckable(true);
     m_actionIconifyFocusLost->setChecked(m_iconifyFocusLost);
     connect(m_actionIconifyFocusLost, SIGNAL(toggled(bool)), this, SLOT(setIconifyFocusLost(bool)));
     m_optionsMenu->addAction(m_actionIconifyFocusLost);
+
+    m_actionIconifyOnClose = new QAction(tr("Iconify on close"), m_optionsMenu);
+    m_actionIconifyOnClose->setCheckable(true);
+    m_actionIconifyOnClose->setChecked(m_iconifyOnClose);
+    connect(m_actionIconifyOnClose, SIGNAL(toggled(bool)), this, SLOT(setIconifyOnClose(bool)));
+    m_optionsMenu->addAction(m_actionIconifyOnClose);
+
     m_actionBalloonTitleChanges = new QAction(tr("Balloon title changes"), m_optionsMenu);
     m_actionBalloonTitleChanges->setCheckable(true);
     m_actionBalloonTitleChanges->setChecked(m_balloonTimeout ? true : false);
     connect(m_actionBalloonTitleChanges, SIGNAL(triggered(bool)), this, SLOT(setBalloonTimeout(bool)));
     m_optionsMenu->addAction(m_actionBalloonTitleChanges);
-    m_contextMenu->addMenu(m_optionsMenu);
 
+    m_contextMenu->addMenu(m_optionsMenu);
     m_contextMenu->addAction(QIcon(":/images/another.png"), tr("Dock Another"), this, SLOT(doSelectAnother()));
     m_contextMenu->addAction(tr("Undock All"), this, SLOT(doUndockAll()));
     m_contextMenu->addSeparator();
@@ -546,7 +566,8 @@ void TrayItem::createContextMenu() {
     connect(m_actionToggle, SIGNAL(triggered()), this, SLOT(toggleWindow()));
     m_contextMenu->addAction(m_actionToggle);
     m_contextMenu->addAction(tr("Undock"), this, SLOT(doUndock()));
-    m_contextMenu->addAction(QIcon(":/images/close.png"), tr("Close"), this, SLOT(close()));
+    m_contextMenu->addAction(QIcon(":/images/close.png"), tr("Close"), this, SLOT(closeWindow()));
+
     setContextMenu(m_contextMenu);
 }
 
@@ -560,6 +581,7 @@ QIcon TrayItem::createIcon(Window window) {
     QPixmap appIcon;
     Display *display = QX11Info::display();
     XWMHints *wm_hints = XGetWMHints(display, window);
+
     if (wm_hints != 0) {
         if (!(wm_hints->flags & IconMaskHint))
             wm_hints->icon_mask = None;
@@ -577,20 +599,10 @@ QIcon TrayItem::createIcon(Window window) {
     if (!window_icon) {
         appIcon.load(":/images/question.png");
     } else {
-        appIcon = QPixmap((const char **) window_icon);
+        appIcon = QPixmap(const_cast<const char **> (window_icon));
     }
     if (window_icon) {
         XpmFree(window_icon);
     }
     return QIcon(appIcon);
 }
-
-bool TrayItem::isBadWindow() {
-    Display *display = QX11Info::display();
-
-    if (!isValidWindowId(display, m_window)) {
-        destroyEvent();
-        return true;
-    }
-    return false;
-}
